module fbi_bm_mod

  use sep
  use mem_mod
  use tomo_mod
  use sscale_mod
  use lint_mod
  use step_mod
  use scats0_mod
  use escats0_mod
  use secder_mod

  implicit none
  integer, private                                     :: nt, n1, n2, nt2, wfld_j
  integer, private                                     :: sou_nx, sou_ox, sou_dx, sou_z, isou, sou_x
  integer, private                                     :: rec_nx, rec_ox, rec_dx, rec_z, irec, rec_x
  integer, private                                     :: nh, oh, it
  integer, private                                     :: bc1, bc2, bc3, bc4
  real, private                                        :: dt, d1, o1, d2, o2, dtw
  real, dimension(:), allocatable, private             :: wavelet
  real, dimension(:,:), allocatable, target, private   :: vel, veldt2, vwt, wt
  real, dimension(:,:,:), allocatable, target, private :: temp1, temp2, temp1i, temp2i
  real, dimension(:,:,:,:), allocatable, target, private :: wflds0
  contains

  subroutine init1(n1_in, o1_in, d1_in, n2_in, o2_in, d2_in, bc1_in, bc2_in, bc3_in, bc4_in)
    integer                             :: n1_in, n2_in, bc1_in, bc2_in, bc3_in, bc4_in
    real                                :: d1_in, o1_in, d2_in, o2_in
    n1 = n1_in
    n2 = n2_in
    d1 = d1_in
    o1 = o1_in
    d2 = d2_in
    o2 = o2_in
    bc1 = bc1_in
    bc2 = bc2_in
    bc3 = bc3_in
    bc4 = bc4_in
  end subroutine

  subroutine init2(sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in, rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in)
    integer                             :: sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in
    integer                             :: rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in
    sou_nx = sou_nx_in
    sou_ox = sou_ox_in
    sou_dx = sou_dx_in
    sou_z = sou_z_in
    rec_nx = rec_nx_in
    rec_ox = rec_ox_in
    rec_dx = rec_dx_in
    rec_z = rec_z_in
  end subroutine

  subroutine init3(nt_in, dt_in, wfld_j_in, nh_in, oh_in)
    integer                             :: nt_in, wfld_j_in, nh_in, oh_in
    real                                :: dt_in
    nt = nt_in
    wfld_j = wfld_j_in
    dt = dt_in
    nh = nh_in
    oh = oh_in
  end subroutine

  subroutine init_alloc(wavelet_in)
    real, dimension(nt)                  :: wavelet_in

    nt2 = (nt-1)*wfld_j+1
    dtw = dt / wfld_j

    call mem_alloc1d(wavelet, nt)
    call mem_alloc2d(vel, n1, n2)
    call mem_alloc2d(veldt2, n1, n2)
    call mem_alloc2d(vwt, n1, n2)
    call mem_alloc2d(wt, n1, n2)
    call mem_alloc4d(wflds0, n1, n2, nt, sou_nx)
    call mem_alloc3d(temp1, n1, n2, nt)
    call mem_alloc3d(temp2, n1, n2, nt)
    call mem_alloc3d(temp1i, n1, n2, nt2)
    call mem_alloc3d(temp2i, n1, n2, nt2)

    wavelet = wavelet_in

    call sscale_init(n1, n2, nt, dtw, veldt2, wt)
    call lint_init(n1, n2, nt, wfld_j)
    call step_init(n1, n2, nt2, bc1, bc2, bc3, bc4, d1, d2, veldt2, vwt, wt)
    call scats0_init(n1, n2, nt, wflds0(:,:,:,1))
    call escats0_init(n1, n2, nt, nh, oh, wflds0(:,:,:,1))
    call secder_init(n1, n2, nt, dt)

  end subroutine

  subroutine model_update(vel_in)
    real, dimension(n1, n2)              :: vel_in
    vel = 0.
    vwt = 0.
    wt = 0.
    vel = vel_in
    call AbcWt(vel, vwt, wt, dtw, d1, d2, bc1, bc2, bc3, bc4)
    veldt2 = vel*vel*dtw*dtw
  end subroutine

  subroutine compute_wflds0()
    do isou=1,sou_nx
      sou_x = (isou-1)*sou_dx + sou_ox
      temp1 = 0.
      temp1(sou_z,sou_x,:) = wavelet
      call sscale_op2(.false., .false., temp1, temp2)
      call lint_op2(.false., .false., temp2, temp1i)
      call step_op2(.false., .false., temp1i, temp2i)
      call lint_op2(.true., .false., wflds0(:,:,:,isou), temp2i)
    end do
  end subroutine

  function L_nlop(adj, add, model, data) result(stat)
    integer             :: stat
    logical, intent(in) :: adj, add 
    real, dimension(:)  :: model, data
    call L_nlop2(adj, add, model, data)
    stat=0
  end function

  function L_lop(adj, add, model, data) result(stat)
    integer             :: stat 
    logical, intent(in) :: adj, add 
    real, dimension(:)  :: model, data
    call L_lop2(adj, add, model, data)
    stat=0
  end function

  subroutine L_nlop2(adj, add, model, data)
    logical, intent(in)                  :: adj, add
    real, dimension(nt, rec_nx, sou_nx)  :: model
    real, dimension(nt, rec_nx, sou_nx)  :: data
    if(adj) then
      if(.not. add) model = 0.
      do isou=1,sou_nx
        temp1 = 0.
        do irec=1,rec_nx
          rec_x = (irec-1)*rec_dx + rec_ox
          temp1(rec_z,rec_x,:) = data(:,irec,isou)
        end do
        call lint_op2(.false., .false., temp1, temp2i)
        call step_op2(.true., .false., temp1i, temp2i)
        call lint_op2(.true., .false., temp2, temp1i)
        call sscale_op2(.true., .false., temp1, temp2)
        do irec=1,rec_nx
          rec_x = (irec-1)*rec_dx + rec_ox
          model(:,irec,isou) = model(:,irec,isou) + temp1(sou_z,rec_x,:)
        end do
      end do
    else
      if(.not. add) data = 0.
      do isou=1,sou_nx
        temp1 = 0.
        do irec=1,rec_nx
          rec_x = (irec-1)*rec_dx + rec_ox
          temp1(sou_z,rec_x,:) = model(:,irec,isou)
        end do
        call sscale_op2(.false., .false., temp1, temp2)
        call lint_op2(.false., .false., temp2, temp1i)
        call step_op2(.false., .false., temp1i, temp2i)
        call lint_op2(.true., .false., temp1, temp2i)
        do irec=1,rec_nx
          rec_x = (irec-1)*rec_dx + rec_ox
          data(:,irec,isou) = data(:,irec,isou) + temp1(rec_z,rec_x,:)
        end do
      end do
    end if
  end subroutine

  subroutine L_lop2(adj, add, model, data)
    logical, intent(in)                  :: adj, add
    real, dimension(n1, n2, nh)          :: model
    real, dimension(nt, rec_nx, sou_nx)  :: data
    if(adj) then
      if(.not. add) model = 0.
      do isou=1,sou_nx
        call escats0_init(n1, n2, nt, nh, oh, wflds0(:,:,:,isou))
        temp1 = 0.
        do irec=1,rec_nx
          rec_x = (irec-1)*rec_dx + rec_ox
          temp1(rec_z,rec_x,:) = data(:,irec,isou)
        end do
        call lint_op2(.false., .false., temp1, temp2i)
        call step_op2(.true., .false., temp1i, temp2i)
        call lint_op2(.true., .false., temp1, temp1i)
        call sscale_op2(.true., .false., temp2, temp1)
        call secder_op2(.true., .false., temp1, temp2)
        call escats0_op2(.true., .true., model, temp1)
      end do
    else
      if(.not. add) data = 0.
      do isou=1,sou_nx
        call escats0_init(n1, n2, nt, nh, oh, wflds0(:,:,:,isou))
        call escats0_op2(.false., .false., model, temp1)
        call secder_op2(.false., .false., temp1, temp2)
        call sscale_op2(.false., .false., temp2, temp1)
        call lint_op2(.false., .false., temp1, temp1i)
        call step_op2(.false., .false., temp1i, temp2i)
        call lint_op2(.true., .false., temp1, temp2i)
        do irec=1,rec_nx
          rec_x = (irec-1)*rec_dx + rec_ox
          data(:,irec,isou) = data(:,irec,isou) + temp1(rec_z,rec_x,:)
        end do
      end do
    end if
  end subroutine

end module 
